import io
import json
import math
import pickle
import uuid
from dataclasses import dataclass
from datetime import datetime, timedelta
from enum import Enum, auto
from typing import Any, Generator

import numpy as np
import pandas as pd
from bloom_filter2 import BloomFilter
from pydruid.utils.aggregators import longsum
from pydruid.utils.dimensions import DimensionSpec, RegexFilteredSpec
from pydruid.utils.filters import Dimension, Filter

import openapi.models.base_anomaly_event_detail  # needed to avoid circular import for any subclasses  # noqa: F401
from common.utils import logging
from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import Interval, PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.detector.core import BatchDetector, DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import (
    Compute,
    ComputeType,
    DataFlow,
    EventInfo,
    EventKey,
    EventMetadata,
    IpMask,
    JsonValue,
    ListString,
    MergePlan,
    MergePolicy,
    Score,
    Tactic,
    Technique,
)
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery, FlowVirtualColumn
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    DetectorRunStatus,
    DetectorStatusLabel,
    ExclusionLabel,
    ExclusionTypeInfo,
    UserConfigInfo,
)
from common.utils.vizfilter.core import Condition, Criteria, VizFilter, VizFilterOption
from llmnrnbtns.detector.config import Config
from openapi.models.name_resolution_report import NameResolutionReport
from openapi.models.series import Series
from vmware.nsx.intelligence.nta.v1 import nta_pb2


class LlmnrNbtnsDetectionError(Exception):
    pass


class AnomalySubtype(Enum):
    LLMNR = 0
    NBTNS = auto()


@dataclass
class PartialBaselineData:
    llmnr_data: io.BytesIO
    nbtns_data: io.BytesIO


@dataclass
class BaselineData:
    llmnr_data: io.BytesIO
    nbtns_data: io.BytesIO
    llmnr_datacenter_data_per_day: dict[datetime, BloomFilter]
    nbtns_datacenter_data_per_day: dict[datetime, BloomFilter]


class LlmnrNbtns(BatchDetector):
    logger = logging.get_common_logger(__name__)
    LLMNR_UDP_PORT = "5355"
    NBT_UDP_PORT = "137"
    DEFAULT_STD = 3
    MAX_SCORE = 100
    DEFAULT_SCORE = 50
    DEFAULT_VM_COUNT = 10
    VM_COUNT_QUADRATIC_COEFFICIENT = DEFAULT_SCORE * DEFAULT_SCORE / DEFAULT_VM_COUNT
    STD_QUADRATIC_COEFFICIENT = DEFAULT_SCORE * DEFAULT_SCORE / DEFAULT_STD
    DEFAULT_SENSITIVITY = 50

    def __init__(
        self,
        baseline_service: BaselineService,
        checkpoint_service: CheckpointService,
        user_config_service: UserConfigService,
        event_service: EventService,
        flow_service: FlowService,
        lookup_service: LookupService,
        now: datetime,
    ):
        super().__init__(baseline_service, checkpoint_service, user_config_service, event_service, now)
        self._flow_service = flow_service
        self._lookup_service = lookup_service
        self._now = now
        self._beginning_of_today = datetime.combine(now.date(), datetime.min.time())

    @staticmethod
    def detector_name() -> DetectorType:
        return DetectorType.LLMNR_NBTNS

    @staticmethod
    def _baseline_max_retention_time() -> timedelta:
        return timedelta(days=float(Config.BASELINE_MAX_RETENTION_DAYS))

    def _generate_baseline(self, site_id: str, interval: Interval) -> PartialBaseline:
        LlmnrNbtns.logger.info(f"generating LlmnrNbtns baselines for interval {interval}")

        data = PartialBaselineData(
            self._get_llmnr_partial_baseline_data(interval, site_id),
            self._get_nbtns_partial_baseline_data(interval, site_id),
        )

        buffer = io.BytesIO()
        pickle.dump(data, buffer)

        return PartialBaseline(interval, buffer)

    def _get_baseline_intervals(self, missing_intervals: list[Interval]) -> list[Interval]:
        return BaselineService.convert_to_daily_intervals_with_trim(
            missing_intervals,
            self._beginning_of_today,
            timedelta(days=float(Config.BASELINE_MAX_HISTORICAL_CREATION_DAYS)),
        )

    def _get_llmnr_partial_baseline_data(self, interval: Interval, site_id: str) -> io.BytesIO:
        request_count_result = self._query_llmnr_request_count(interval, site_id).dropna()
        if len(request_count_result.index) == 0:
            LlmnrNbtns.logger.info(
                f"got empty request count result when constructing llmnr partial baseline for {interval.start_time}/{interval.end_time}, returning empty partial baseline data"
            )
            buffer = io.BytesIO()
            pd.DataFrame().to_parquet(buffer)
            return buffer

        baseline_result = request_count_result.groupby("dstVmId")[["count"]].sum()
        baseline_result = baseline_result.rename(columns={"count": "request_received_count"})

        request_source_port_result = self._query_llmnr_request_src_port(interval, site_id)
        if len(request_source_port_result.index) == 0:
            LlmnrNbtns.logger.info(
                "got empty request_source_port_result, returning baseline with empty response_count and response_vms"
            )
            baseline_result["response_count"] = np.NaN
            baseline_result["response_vms"] = None
            baseline_result.index.name = "vm_id"
            buffer = io.BytesIO()
            baseline_result.to_parquet(buffer)
            return buffer

        grouped_source_port_result = pd.DataFrame(
            request_source_port_result.groupby("srcVmId")["originalSrcPorts"].apply(list)
        )

        response_result = self._query_llmnr_response(
            interval, site_id, request_source_port_result["originalSrcPorts"].drop_duplicates().to_list()
        )
        if len(response_result.index) == 0:
            LlmnrNbtns.logger.info(
                "got empty response_result, returning baseline with empty response_count and response_vms"
            )
            baseline_result["response_count"] = np.NaN
            baseline_result["response_vms"] = None
            baseline_result.index.name = "vm_id"
            buffer = io.BytesIO()
            baseline_result.to_parquet(buffer)
            return buffer

        # proving, the src vm actually received an llmnr request
        response_result = response_result[response_result["srcVmId"].isin(baseline_result.index)]

        # make sure the dst vm had sent a request for this src port
        response_result = response_result.join(grouped_source_port_result, on="dstVmId", how="inner")

        response_result = response_result[
            response_result.apply(lambda row: row["dstPort"] in row["originalSrcPorts"], axis=1)
        ]

        response_result = response_result.groupby("srcVmId")[["count", "dstVmId"]].agg(
            {"count": "sum", "dstVmId": lambda x: list(x)}
        )

        baseline_result["response_count"] = response_result["count"]
        baseline_result["response_vms"] = response_result["dstVmId"]
        baseline_result.index.name = "vm_id"

        """llmnr partial baseline dataframe:
                    request_received_count      response_count      response_vms
        vm_id
        vm1         44                          523                 ["vm10", "vm20"]
        vm2         543                         23                  ["vm30", "vm40"]
        vm3         8738                        74                  ["vm9", "vm10"]
        vm4         294                         NaN                 None
        ...
        """

        buffer = io.BytesIO()
        baseline_result.to_parquet(buffer)
        return buffer

    def _query_llmnr_request_count(self, interval: Interval, site_id: str) -> pd.DataFrame:
        filters = [
            Dimension("destIpType") == "MCAST",
            Dimension("protocol") == "UDP",
            Dimension("dstPort") == LlmnrNbtns.LLMNR_UDP_PORT,
            Dimension("dstSiteId") == site_id,
            Dimension("isDstInfoPresent") == True,  # noqa
        ]

        dimensions = [DimensionSpec("dstVmId", "dstVmId", filter_spec=RegexFilteredSpec(".*-.*-.*-.*"))]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[interval],
            aggregations={"count": longsum("dstFwAllowed")},
            dimensions=dimensions,
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _query_llmnr_request_src_port(self, interval: Interval, site_id: str) -> pd.DataFrame:
        filters = [
            Dimension("destIpType") == "MCAST",
            Dimension("protocol") == "UDP",
            Dimension("dstPort") == LlmnrNbtns.LLMNR_UDP_PORT,
            Dimension("srcSiteId") == site_id,
            Dimension("isSrcInfoPresent") == True,  # noqa
        ]

        dimensions = [
            DimensionSpec("srcVmId", "srcVmId", filter_spec=RegexFilteredSpec(".*-.*-.*-.*")),
            DimensionSpec("originalSrcPorts", "originalSrcPorts"),
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[interval],
            aggregations={},
            dimensions=dimensions,
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query)

    def _query_llmnr_response(self, interval: Interval, site_id: str, request_source_ports: list[str]) -> pd.DataFrame:
        filters = [
            Dimension("protocol") == "UDP",
            Filter(type="in", dimension="dstPort", values=request_source_ports),
            Dimension("srcSiteId") == site_id,
        ]

        dimensions = [
            DimensionSpec("presentSrcVmId", "srcVmId", filter_spec=RegexFilteredSpec(".*-.*-.*-.*")),
            DimensionSpec("presentDstVmId", "dstVmId", filter_spec=RegexFilteredSpec(".*-.*-.*-.*")),
            DimensionSpec("dstPort", "dstPort"),
        ]

        virtual_columns = [
            FlowVirtualColumn(
                "flowCount",
                "if ( like({0}, 'true') && like({1}, 'true'), {2}+{3}+{4},  if ( like({0}, 'true'), {2}+{5}, {3}+{4}))".format(
                    "isSrcInfoPresent",
                    "isDstInfoPresent",
                    "srcFwBlocked",
                    "dstFwBlocked",
                    "dstFwAllowed",
                    "srcFwAllowed",
                ),
            ),
            self._flow_service.get_present_src_id_virtual_column(),
            self._flow_service.get_present_dst_id_virtual_column(),
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[interval],
            aggregations={"count": longsum("flowCount")},
            dimensions=dimensions,
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
            virtual_columns=virtual_columns,
        )

        return self._flow_service.group_by(query)

    def _get_nbtns_partial_baseline_data(self, interval: Interval, site_id: str) -> io.BytesIO:
        nbtns_flows = self._query_nbtns_flows(interval, site_id).dropna()

        """
                dstVmId     srcVmId     dstIps      dstFwAllowedSum     totalFlowCount  timestamp
        index
        0       vm1         vm10        192.168.0.1 204                 392             ...
        """

        nsx_managed_ips = self._query_nbtns_valid_ips()
        baseline_result = pd.DataFrame()

        # those without add to request_received_count, those with add to response_count and response_vm
        # we can determine that if the dstIps is managed by NSX, it's a response due to the fact that NBTNS uses a broadcast IP
        if len(nbtns_flows.index) > 0:
            # response_result: srcVmId -> RESPONSE_COUNT: totalFlowCount
            # response_result: srcVmId -> RESPONSE_VMS: [dstVMId...]
            # src, attacker ---UCAST---> dst, victim
            response_result = nbtns_flows[nbtns_flows["dstIps"].isin(nsx_managed_ips)]
            baseline_result = response_result.groupby("srcVmId")[["dstVmId", "totalFlowCount"]].agg(
                {"dstVmId": lambda x: list(x), "totalFlowCount": "sum"}
            )
            baseline_result = baseline_result.rename(
                columns={"dstVmId": "response_vms", "totalFlowCount": "response_count"}
            )

            # request_result: dstVmId -> REQUEST_RECEIVED_COUNT: dstFwAllowedSum
            # src, victim ---BCAST---> dst, attacker
            request_result = nbtns_flows[~nbtns_flows["dstIps"].isin(nsx_managed_ips)]
            baseline_result = baseline_result.join(
                request_result.groupby("dstVmId")["dstFwAllowedSum"].sum(), how="outer"
            )
            baseline_result = baseline_result.rename(columns={"dstFwAllowedSum": "request_received_count"})
            baseline_result.index.name = "vm_id"

        """nbtns partial baseline dataframe:
                response_vms     response_count     request_received_count
        vm_id
        vm1       [vm1, vm2]        82.0             323.0
        vm2       None              NaN              23.0
        vm3       [vm1, vm2]        32.0             NaN
        """

        buffer = io.BytesIO()
        baseline_result.to_parquet(buffer)
        return buffer

    def _query_nbtns_flows(self, interval: Interval, site_id: str) -> pd.DataFrame:
        filters = [
            Dimension("protocol") == "UDP",
            Dimension("dstPort") == LlmnrNbtns.NBT_UDP_PORT,
            Dimension("dstSiteId") == site_id,
        ]

        virtual_columns = [
            FlowVirtualColumn(
                "totalFlowCount",
                "if ( like({0}, 'true') && like({1}, 'true'), {2}+{3}+{4},  if ( like({0}, 'true'), {2}+{5}, {3}+{4}))".format(
                    "isSrcInfoPresent",
                    "isDstInfoPresent",
                    "srcFwBlocked",
                    "dstFwBlocked",
                    "dstFwAllowed",
                    "srcFwAllowed",
                ),
            ),
            self._flow_service.get_present_src_id_virtual_column(),
            self._flow_service.get_present_dst_id_virtual_column(),
        ]

        dimensions = [
            DimensionSpec("presentSrcVmId", "srcVmId"),
            DimensionSpec("presentDstVmId", "dstVmId"),
            "dstIps",
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[interval],
            aggregations={"dstFwAllowedSum": longsum("dstFwAllowed"), "totalFlowCount": longsum("totalFlowCount")},
            dimensions=dimensions,
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
            virtual_columns=virtual_columns,
        )

        return self._flow_service.group_by(query)

    def _query_nbtns_valid_ips(self) -> pd.Series:
        interval = Interval(start_time=datetime(year=1970, month=1, day=1, hour=1, minute=1), end_time=self._now)

        query = FlowGroupByQuery(
            datasource=Config.IPS_DATA_SOURCE,
            granularity=Config.IPS_QUERY_GRANULARITY,
            intervals=[interval],
            aggregations={},
            dimensions=["ip_addresses"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Dimension("config_type") == "VM",
        )

        result = self._flow_service.group_by(query).dropna()
        if len(result.index) == 0:
            return pd.Series()

        return result["ip_addresses"]

    def _combine_baseline_objects(self, baselines: list[PartialBaseline]) -> io.BytesIO | None:
        if len(baselines) == 0:
            LlmnrNbtns.logger.warn("got empty list of baselines to combine, returning empty baseline for detection")
            return None
        LlmnrNbtns.logger.info(f"got {len(baselines)} partial baselines to merge from the baseline store")

        llmnr_baseline_data = pd.DataFrame()
        nbtns_baseline_data = pd.DataFrame()
        for partial_baseline in baselines:
            if partial_baseline.interval.end_time > self._beginning_of_today:
                LlmnrNbtns.logger.info(
                    f"skipping baseline with endtime of {partial_baseline.interval.end_time} as it is considered too current"
                )
                continue

            partial_baseline.data.seek(0)
            partial_baseline_data: PartialBaselineData = pickle.load(partial_baseline.data)

            partial_llmnr_data = pd.read_parquet(partial_baseline_data.llmnr_data)
            if len(partial_llmnr_data.index) > 0:
                partial_llmnr_data["start_time"] = partial_baseline.interval.start_time
                partial_llmnr_data = partial_llmnr_data.set_index(["start_time"], append=True)
                llmnr_baseline_data = pd.concat([llmnr_baseline_data, partial_llmnr_data])

            partial_nbtns_data = pd.read_parquet(partial_baseline_data.nbtns_data)
            if len(partial_nbtns_data.index) > 0:
                partial_nbtns_data["start_time"] = partial_baseline.interval.start_time
                partial_nbtns_data = partial_nbtns_data.set_index("start_time", append=True)
                nbtns_baseline_data = pd.concat([nbtns_baseline_data, partial_nbtns_data])

        llmnr_baseline_data = llmnr_baseline_data.sort_index()
        nbtns_baseline_data = nbtns_baseline_data.sort_index()

        """
                              response_vms     response_count     request_received_count
        vm_id   start_time
        vm1     2023-01-01    [vm10, vm20, vm30]  47.0               281.0
                2023-01-02    [vm20]              32.0               444.0
                2023-01-03    [vm30, vm20]        2.0                39.0
                2023-01-04    None                NaN                39.0
        vm2     2023-01-02    [vm10]              32.0               72.0
                2023-01-03    [vm30, vm20]        2.0                39.0
        """

        llmnr_buffer = io.BytesIO()
        nbtns_buffer = io.BytesIO()

        llmnr_baseline_data.to_parquet(llmnr_buffer)
        nbtns_baseline_data.to_parquet(nbtns_buffer)

        baseline_data = BaselineData(
            llmnr_buffer,
            nbtns_buffer,
            self._build_datacenter_baseline(llmnr_baseline_data),
            self._build_datacenter_baseline(nbtns_baseline_data),
        )

        baseline_buffer = io.BytesIO()
        pickle.dump(baseline_data, baseline_buffer)
        return baseline_buffer

    def _build_datacenter_baseline(self, baseline: pd.DataFrame) -> dict[datetime, BloomFilter]:
        datacenter_data_per_day: dict[datetime, BloomFilter] = {}
        for index, row in baseline.iterrows():
            if row["response_vms"] is None:
                continue

            compute_id = index[0]  # type: ignore
            start_time = index[1]  # type: ignore

            if start_time not in datacenter_data_per_day:
                bloom_filter = BloomFilter(max_elements=100000, error_rate=0.001)
                datacenter_data_per_day[start_time] = bloom_filter

            for responded_vm in row["response_vms"]:
                datacenter_data_per_day[start_time].add(f"{compute_id}:{responded_vm}")

        return datacenter_data_per_day

    def _build_datacenter_data_for_today(self, data: pd.DataFrame) -> BloomFilter:
        datacenter_data: BloomFilter = BloomFilter(max_elements=100000, error_rate=0.001)
        for compute_id, row in data.iterrows():
            if row["response_vms"] is None:
                continue

            for responded_vm in row["response_vms"]:
                datacenter_data.add(f"{compute_id}:{responded_vm}")

        return datacenter_data

    def _detect_events(
        self, user_config: UserConfigInfo, baseline: io.BytesIO, checkpoint: io.BytesIO | None
    ) -> Generator[EventInfo, Any, Any]:
        # query for today's data
        current_data_buffer = self._generate_baseline(
            user_config.site_id, Interval(self._beginning_of_today, self._now)
        ).data

        # deser today's data
        current_data_buffer.seek(0)
        current_data: PartialBaselineData = pickle.load(current_data_buffer)
        current_llmnr_data = pd.read_parquet(current_data.llmnr_data)
        current_nbtns_data = pd.read_parquet(current_data.nbtns_data)

        # deser baseline
        baseline.seek(0)
        baseline_data: BaselineData = pickle.load(baseline)
        baseline_llmnr_data = pd.read_parquet(baseline_data.llmnr_data)
        baseline_nbtns_data = pd.read_parquet(baseline_data.nbtns_data)

        # detect events
        LlmnrNbtns.logger.info("starting detection for llmnr")
        for event in self._detect_events_on(
            user_config,
            current_llmnr_data,
            baseline_llmnr_data,
            baseline_data.llmnr_datacenter_data_per_day,
            AnomalySubtype.LLMNR,
        ):
            LlmnrNbtns.logger.info("detected event for llmnr")
            yield event

        LlmnrNbtns.logger.info("starting detection for nbtns")
        for event in self._detect_events_on(
            user_config,
            current_nbtns_data,
            baseline_nbtns_data,
            baseline_data.nbtns_datacenter_data_per_day,
            AnomalySubtype.NBTNS,
        ):
            LlmnrNbtns.logger.info("detected event for nbtns")
            yield event

    def _detect_events_on(
        self,
        user_config: UserConfigInfo,
        computes_detecting_on: pd.DataFrame,
        baseline_data: pd.DataFrame,
        datacenter_data_per_day: dict[datetime, BloomFilter],
        anomaly_subtype: AnomalySubtype,
    ) -> Generator[EventInfo, Any, Any]:
        """computes detecting on DataFrame:
                response_vms     response_count     request_received_count
        vm_id
        vm1       [vm1, vm2]        82.0             323.0
        vm2       None              NaN              23.0
        vm3       [vm1, vm2]        32.0             NaN
        """

        """baseline data DataFrame:
                              response_vms     response_count     request_received_count
        vm_id   start_time
        vm1     2023-01-01    [vm10, vm20, vm30]  47.0               281.0
                2023-01-02    [vm20]              32.0               444.0
                2023-01-03    [vm30, vm20]        2.0                39.0
                2023-01-04    None                NaN                39.0
        vm2     2023-01-02    [vm10]              32.0               72.0
                2023-01-03    [vm30, vm20]        2.0                39.0


        NumberOfResponses == sum(RESPONSE_COUNT)
        NumberOfRequests == sum(REQUEST_RECEIVED_COUNT)
        NumberOfVmsResponded == sum(RESPONSE_VMS)
        """

        # pre-flight checks
        if len(baseline_data.index) == 0:
            LlmnrNbtns.logger.info("baseline is empty, skipping detection")
            return

        if len(computes_detecting_on.index) == 0:
            LlmnrNbtns.logger.info("detection data for today is empty, skipping detection")
            return

        all_vms = self._get_vms_from_flows(user_config.site_id)
        if len(all_vms.index) == 0:
            LlmnrNbtns.logger.info("found no VMs from flows, skipping detection")
            return

        # drop excluded VMs
        excluded_computes = self._get_excluded_computes_from_config(user_config)
        if LlmnrNbtns.logger.getEffectiveLevel() == logging.DEBUG:
            LlmnrNbtns.logger.debug(f"got {len(excluded_computes)} computes from the exclusion list")
            LlmnrNbtns.logger.debug(f"computes to exclude: {excluded_computes}")
        computes_detecting_on = computes_detecting_on[~computes_detecting_on.index.isin(excluded_computes)]

        # check the age of the baseline
        min_baseline_age = timedelta(days=int(Config.MIN_BASELINE_DURATION_DAYS))
        baseline_age = self._now - baseline_data.index.get_level_values(1).min()
        if min_baseline_age > baseline_age:
            LlmnrNbtns.logger.info(
                f"baseline is not old enough to be used for detection, age is {baseline_age} and minimum age is {min_baseline_age}"
            )

        # faking baselines for new computes with beginning_of_today timestamp
        new_computes = computes_detecting_on[
            ~computes_detecting_on.index.isin(baseline_data.index.get_level_values(0))
        ].index
        new_computes_index = pd.MultiIndex.from_frame(
            pd.DataFrame([[compute_name, self._beginning_of_today] for compute_name in new_computes]),
            names=baseline_data.index.names,
        )
        # For new computes, baseline data does not exist
        new_computes_baseline = pd.DataFrame(
            [[None, np.NaN, np.NaN]], index=new_computes_index, columns=baseline_data.columns
        )
        baseline_data = pd.concat([baseline_data, new_computes_baseline])

        # drop baseline rows which we're not detecting on
        baseline_data = baseline_data[baseline_data.index.isin(computes_detecting_on.index, level=0)]

        if len(baseline_data.index) == 0:
            LlmnrNbtns.logger.info("baseline is empty after filtering out invalid baselines, skipping detection")
            return

        # drop computes from detection if they don't have a baseline because they're excluded above
        computes_detecting_on = computes_detecting_on[
            computes_detecting_on.index.isin(baseline_data.index.get_level_values(0))
        ]

        # clean up baseline missing data, interpret missing data (NaN) as 0
        baseline_data[["response_count", "request_received_count"]] = baseline_data[
            ["response_count", "request_received_count"]
        ].replace(np.NaN, 0)

        # clean up computes detecting on, interpret missing data (NaN) as 0
        computes_detecting_on[["response_count", "request_received_count"]] = computes_detecting_on[
            ["response_count", "request_received_count"]
        ].replace(np.NaN, 0)

        # sort the dataframes for comparison
        baseline_data = baseline_data.sort_index()
        computes_detecting_on = computes_detecting_on.sort_index()

        if len(computes_detecting_on.index) == 0:
            LlmnrNbtns.logger.info("no baseline data for the computes found in today's flows, skipping detection")
            return

        # calculate the response count threshold for each VM
        # if above the threshold, calculate STD score, else STD score defaults to 0
        computes_detecting_on["std_score"] = 0

        response_count_thresholds = self._calculate_thresholds(baseline_data["response_count"])
        request_to_response_ratio_thresholds = self._calculate_request_to_response_ratio_thresholds(baseline_data)

        computes_detecting_on["response_to_request_ratio"] = np.NaN
        computes_detecting_on.loc[computes_detecting_on["request_received_count"] > 0, "response_to_request_ratio"] = (
            computes_detecting_on["response_count"] / computes_detecting_on["request_received_count"]
        )

        # the compute hasn't received any domain resolution requests that we can tell (usually due to it being an external IP), assume it's replying to every request
        computes_detecting_on["response_to_request_ratio"] = computes_detecting_on["response_to_request_ratio"].replace(
            np.NaN, int(Config.DEFAULT_RESPONSE_TO_REQUEST_RATIO)
        )

        score_values = self._calculate_std_score_values(computes_detecting_on, baseline_data, "response_count")
        computes_detecting_on.loc[
            (computes_detecting_on["response_count"] > response_count_thresholds)
            & (computes_detecting_on["response_to_request_ratio"] > request_to_response_ratio_thresholds),
            "std_score",
        ] = self._calculate_quadratic_scores(score_values, LlmnrNbtns.STD_QUADRATIC_COEFFICIENT)

        # calculate VM count threshold
        # if above the threshold, calculate STD VM COUNT score, else defaults to 0
        computes_detecting_on["std_vm_count_score"] = 0
        baseline_data["number_of_vms_responded"] = baseline_data["response_vms"].str.len().replace(np.NaN, 0)
        computes_detecting_on["number_of_vms_responded"] = (
            computes_detecting_on["response_vms"].str.len().replace(np.NaN, 0)
        )

        vm_count_thresholds = self._calculate_thresholds(baseline_data["number_of_vms_responded"])

        score_values = self._calculate_std_score_values(computes_detecting_on, baseline_data, "number_of_vms_responded")
        computes_detecting_on.loc[
            computes_detecting_on["number_of_vms_responded"] > vm_count_thresholds, "std_vm_count_score"
        ] = self._calculate_quadratic_scores(score_values, LlmnrNbtns.STD_QUADRATIC_COEFFICIENT)

        # build datacenter data for today
        current_datacenter_data = self._build_datacenter_data_for_today(computes_detecting_on)

        # if current weighted score is > 0, get unique response VMs
        # calculate VM uniqueness score for each
        computes_detecting_on["vm_uniqueness_score"] = 0
        computes_detecting_on["num_vms_never_responded_before"] = np.NaN
        for index, row in computes_detecting_on.iterrows():
            if row["std_score"] == 0 and row["std_vm_count_score"] == 0:
                continue

            compute_id: str = index  # type: ignore

            potential_vms_responded_today = self._get_potential_responders(compute_id, all_vms, current_datacenter_data)
            num_vms_never_responded_before = self._get_num_vms_never_responded_before(
                potential_vms_responded_today, datacenter_data_per_day
            )
            computes_detecting_on.loc[compute_id, "num_vms_never_responded_before"] = num_vms_never_responded_before

            score = self._calculate_quadratic_score(
                num_vms_never_responded_before, LlmnrNbtns.VM_COUNT_QUADRATIC_COEFFICIENT
            )
            computes_detecting_on.loc[compute_id, "vm_uniqueness_score"] = score

        # calculate the final weighted score, confidence, and compare to sensitivity
        # if is silent or not, and confident enough or not, create event
        computes_detecting_on["confidence"] = self._calculate_confidence(computes_detecting_on)

        default_max_columns = pd.options.display.max_columns
        pd.options.display.max_columns = 5
        LlmnrNbtns.logger.info(
            f"confidence and score for each compute:\n{computes_detecting_on[['std_score', 'std_vm_count_score', 'vm_uniqueness_score', 'confidence']]}"
        )
        pd.options.display.max_columns = default_max_columns

        # if not reporting silent events, drop all rows with confidence < sensitivity
        LlmnrNbtns.logger.info(f"{user_config.settings}")
        sensitivity_setting = user_config.settings.get_sensitivity_meter_setting()
        if sensitivity_setting is None:
            raise LlmnrNbtnsDetectionError("could not read sensitivity setting from user config store, got None")
        if Config.PERSIST_LOW_CONFIDENCE_EVENTS_AS_SILENT != "True":
            computes_detecting_on = computes_detecting_on[
                computes_detecting_on["confidence"] >= sensitivity_setting.sensitivity
            ]
            LlmnrNbtns.logger.info(
                f"after dropping low confidence events, num events to report is {len(computes_detecting_on.index)}"
            )

        computes_detecting_on["is_silent"] = computes_detecting_on["confidence"] < sensitivity_setting.sensitivity

        # prepare a compute->ip lookup table for all events
        compute_to_ip_lookup = self._lookup_service.computes_to_ip_lookup(
            user_config.site_id, Interval(self._beginning_of_today, self._now)
        )
        LlmnrNbtns.logger.info(f"preparing compute to ip lookup (size:{len(compute_to_ip_lookup)})")

        # create events to report
        for index, row in computes_detecting_on.iterrows():
            compute_id = str(index)
            LlmnrNbtns.logger.info(f"creating event for compute: {compute_id}")
            potential_vms_responded_today = self._get_potential_responders(compute_id, all_vms, current_datacenter_data)
            num_new_vm_responses = self._get_num_vms_never_responded_before(
                potential_vms_responded_today, datacenter_data_per_day
            )
            current_response_count = row["response_count"] if not pd.isna(row["response_count"]) else 0
            current_response_vms_count = len(row["response_vms"]) if row["response_vms"] is not None else 0
            yield self._create_event(
                compute_id,
                row,
                anomaly_subtype,
                potential_vms_responded_today,
                num_new_vm_responses,
                baseline_data.loc[compute_id]["response_count"].to_list() + [current_response_count],
                baseline_data.loc[compute_id]["response_vms"].str.len().replace(np.NaN, 0).to_list()
                + [current_response_vms_count],
                baseline_data.loc[compute_id].index.to_list() + [self._now],
                user_config.site_id,
                compute_to_ip_lookup,
            )

    def _get_excluded_computes_from_config(self, config: UserConfigInfo) -> list[str]:
        if config.exclusion_settings == {}:
            return []

        excluded_computes = []
        for setting in config.exclusion_settings.exclusions:
            # technically, it should be ExclusionLabel.DESTINATION but the existing API
            # is using NONE. In order to avoid changing the API contract, we allow both
            # DESTINATION & NONE (existing behavior)
            if setting.label == ExclusionLabel.DESTINATION or setting.label == ExclusionLabel.NONE:
                # setting.excluded_items_info.value can technically be any type, but here we're assuming it's a list of str
                excluded_entity_ids: list[str] = setting.excluded_items_info.value  # type: ignore (suppress linter warning)
                if setting.type == ExclusionTypeInfo.VM:
                    excluded_computes += excluded_entity_ids
                elif setting.type == ExclusionTypeInfo.GROUP:
                    excluded_computes += self._lookup_service.groups_to_computes_lookup(
                        config.site_id, excluded_entity_ids
                    )
        return excluded_computes

    def _get_vms_from_flows(self, site_id: str) -> pd.DataFrame:
        filters = [
            Dimension("srcComputeType") == "VM",
            Dimension("srcSiteId") == site_id,
            Dimension("isSrcInfoPresent") == True,  # noqa
        ]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity="all",
            intervals=[Interval(start_time=self._now - self._baseline_max_retention_time(), end_time=self._now)],
            aggregations={"lastStartTime": {"type": "longFirst", "fieldName": "lastStartTime"}},
            dimensions=["srcVmId"],
            context={"queryId": f"NTA:{self.detector_name()}:{uuid.uuid4()}"},
            filter=Filter(type="and", fields=filters),
        )

        return self._flow_service.group_by(query).dropna()

    def _calculate_thresholds(self, nums: pd.Series) -> pd.Series:
        """Calculate the threshold used to determine if we should score each compute, score is later used to calculate confidence
        For computes that don't have a baseline (responders that have not received any request or responded in the past), we set
        the std to 0, since the implied baseline is they have 0 for each statistic. The result is the threshold for these computes
        is 0. Since it's unexpected that a new compute will suddenly respond to requests without a baseline, this threshold is ideal.
        """
        stats = nums.groupby("vm_id").agg(["mean", "std"])

        # if there's no baseline for this compute, replace it with 0, threshold should be 0
        stats["std"] = stats["std"].replace(np.NaN, 0)

        stats["upper_bound"] = stats["mean"] + (LlmnrNbtns.DEFAULT_STD * stats["std"])
        stats["lower_bound"] = stats["mean"] - (LlmnrNbtns.DEFAULT_STD * stats["std"])

        filtered_nums = pd.DataFrame(nums)
        val_key = filtered_nums.columns[0]
        filtered_nums = filtered_nums.join(stats[["upper_bound", "lower_bound", "std"]])
        filtered_nums = filtered_nums[
            (filtered_nums["std"] == 0)
            | (
                (filtered_nums[val_key] > filtered_nums["lower_bound"])
                & (filtered_nums[val_key] < filtered_nums["upper_bound"])
            )
        ][val_key]

        adjusted_stats = filtered_nums.groupby("vm_id").agg(["mean", "std"])

        # if there's no baseline for this compute, replace it with 0, threshold should be 0
        adjusted_stats["std"] = adjusted_stats["std"].replace(np.NaN, 0)

        return adjusted_stats["mean"] + (LlmnrNbtns.DEFAULT_STD * adjusted_stats["std"])

    def _calculate_request_to_response_ratio_thresholds(self, baseline_data: pd.DataFrame) -> pd.Series:
        """Calculate the threshold used to determine if we should score each compute, score is later used to calculate confidence
        For computes that don't have a baseline (responders that have not received any request or responded in the past), we set
        the std to 0, since the implied baseline is they have 0 for each statistic. The result is the threshold for these computes
        is 0. Since it's unexpected that a new compute will suddenly respond to requests without a baseline, this threshold is ideal.
        """
        stats = baseline_data.groupby("vm_id")["response_count"].agg(["mean", "std"])

        # if there's no baseline for this compute, replace it with 0, threshold should be 0
        stats["std"] = stats["std"].replace(np.NaN, 0)

        stats["upper_bound"] = stats["mean"] + (LlmnrNbtns.DEFAULT_STD * stats["std"])
        stats["lower_bound"] = stats["mean"] - (LlmnrNbtns.DEFAULT_STD * stats["std"])

        baseline_data_without_outliers = baseline_data[["response_count", "request_received_count"]].join(
            stats[["upper_bound", "lower_bound", "std"]]
        )
        baseline_data_without_outliers = baseline_data_without_outliers[
            (baseline_data_without_outliers["std"] == 0)
            | (
                (baseline_data_without_outliers["response_count"] > baseline_data_without_outliers["lower_bound"])
                & (baseline_data_without_outliers["response_count"] < baseline_data_without_outliers["upper_bound"])
            )
        ][["response_count", "request_received_count"]]

        baseline_data_without_outliers["response_to_request_ratio"] = 0
        baseline_data_without_outliers.loc[
            baseline_data_without_outliers["request_received_count"] > 0, "response_to_request_ratio"
        ] = baseline_data_without_outliers["response_count"] / baseline_data_without_outliers["request_received_count"]

        return baseline_data_without_outliers["response_to_request_ratio"].groupby("vm_id").mean()

    def _calculate_std_score_values(
        self, computes_detecting_on: pd.DataFrame, baseline_data: pd.DataFrame, key: str
    ) -> pd.Series:
        """Calculate the Z-Score fore each score, scores are later used to calculate confidence
        For computes that don't have a baseline (responders that have not received any request), we set
        the std in the Z-Score calculation to be the mean of all the other compute's std. In the case
        that there's no other computes to get a mean of std from, we set the std to a default value.

        The idea is if this is a new responder, we can use the average responder std for the whole datacenter
        as a baseline for this responder.
        """
        stats = baseline_data.groupby("vm_id")[key].agg(["mean", "std"])

        # if there's no baseline for this compute, replace it with the average standard deviation for all computes
        stats["std"] = stats["std"].replace([np.NaN, 0], stats["std"].mean())

        # if there's no baseline at all, we assume the default std by setting each row to 1 (gets multiplied by DEFAULT_STD in next step)
        stats["std"] = stats["std"].replace([np.NaN, 0], 1)

        stats["upper_bound"] = stats["mean"] + (LlmnrNbtns.DEFAULT_STD * stats["std"])
        stats["lower_bound"] = stats["mean"] - (LlmnrNbtns.DEFAULT_STD * stats["std"])

        baseline_data_without_outliers = baseline_data[[key]].join(stats[["upper_bound", "lower_bound", "std"]])
        baseline_data_without_outliers = baseline_data_without_outliers[
            (baseline_data_without_outliers["std"] == 0)
            | (
                (baseline_data_without_outliers[key] > baseline_data_without_outliers["lower_bound"])
                & (baseline_data_without_outliers[key] < baseline_data_without_outliers["upper_bound"])
            )
        ][key]

        baseline_data_without_outliers = baseline_data_without_outliers.groupby("vm_id").agg(["mean", "std"])

        # if there's no baseline for this compute, replace it with the average standard deviation for all computes
        baseline_data_without_outliers["std"] = baseline_data_without_outliers["std"].replace(
            [np.NaN, 0], baseline_data_without_outliers["std"].mean()
        )

        # if there's no baseline at all, we assume use a std of 1 as a baseline
        baseline_data_without_outliers["std"] = baseline_data_without_outliers["std"].replace([np.NaN, 0], 1)

        # assume 1-1 match between keys since check was done earlier in the parent method
        baseline_data_without_outliers = computes_detecting_on.join(baseline_data_without_outliers)[
            [key, "mean", "std"]
        ]

        baseline_data_without_outliers["score"] = 0
        baseline_data_without_outliers.loc[baseline_data_without_outliers["std"] != 0, "score"] = (
            computes_detecting_on[key]
            .sub(baseline_data_without_outliers["mean"])
            .div(baseline_data_without_outliers["std"])
        )

        return baseline_data_without_outliers["score"]

    def _calculate_quadratic_scores(self, vals: pd.Series, coeff: float) -> pd.Series:
        return vals.mul(coeff).pow(1.0 / 2.0).clip(upper=LlmnrNbtns.MAX_SCORE)

    def _calculate_quadratic_score(self, val: float, coeff: float) -> float:
        return min(math.sqrt(coeff * val), LlmnrNbtns.MAX_SCORE)

    def _get_potential_responders(
        self, vm_id: str, all_vms: pd.DataFrame, current_datacenter_data: BloomFilter
    ) -> list[str]:
        potential_responders = []
        for _, row in all_vms.iterrows():
            if f"{vm_id}:{row['srcVmId']}" in current_datacenter_data:
                potential_responders.append(row["srcVmId"])

        return potential_responders

    def _get_num_vms_never_responded_before(
        self, potential_vms_responded_today: list[str], datacenter_data_per_day: dict[datetime, BloomFilter]
    ) -> int:
        num_vms_never_responded_before = 0
        for potential_vm in potential_vms_responded_today:
            responder_found_in_the_past = False
            for _, datacenter_data in datacenter_data_per_day.items():
                if potential_vm in datacenter_data:
                    responder_found_in_the_past = True
                    break
            if not responder_found_in_the_past:
                num_vms_never_responded_before += 1

        return num_vms_never_responded_before

    def _calculate_confidence(self, computes_detecting_on: pd.DataFrame) -> pd.Series:
        STD_SCORE_WEIGHT = 0.45
        STD_VM_COUNT_SCORE_WEIGHT = 0.45
        VM_UNIQUENESS_SCORE_WEIGHT = 0.1
        MAX_TOTAL_CONFIDENCE = 100

        return (
            computes_detecting_on["std_score"].mul(STD_SCORE_WEIGHT)
            + computes_detecting_on["std_vm_count_score"].mul(STD_VM_COUNT_SCORE_WEIGHT)
            + computes_detecting_on["vm_uniqueness_score"].mul(VM_UNIQUENESS_SCORE_WEIGHT)
        ).clip(upper=MAX_TOTAL_CONFIDENCE)

    def _create_event(
        self,
        compute_id: str,
        compute_row: pd.Series,
        anomaly_subtype: AnomalySubtype,
        potential_vms_responded_today: list[str],
        num_new_vm_responses: int,
        response_counts: list[float],
        responded_vm_counts: list[int],
        counts_labels: list[pd.Timestamp],
        site_id: str,
        compute_to_ip_lookup: dict[str, str],
    ) -> EventInfo:
        """Creates an event that will be sent to the event-store for processing

        compute_id: id of the compute who generated this event
        compute_row: row of the computes_detecting_on table associated with this compute, holds metadata such as scores
        anomaly_subtype: which subtype generated this detection, such as LLMNR or NBTNS
        potential_vms_responded_today: list of compute_ids that were found in today's datacenter object paired with the above compute_id
        num_new_vm_responses: number of vms in the above list who do not appear in the baseline datacenter object
        response_counts: list of "response_count" field for the whole corpus (baseline and today)
        responded_vm_counts: list of len("response_vms") field for the whole corpus
        counts_labels: timestamps ordered by time, same ordering as the above two lists. these three lists together are used to construct the event detail
        site_id: the unique ID of the site we're detecting events on
        """

        # truncate potential responder VMs, if scale is large we the event can get rejected by kafka for being too large
        potential_vms_responded_today = potential_vms_responded_today[: int(Config.EVENT_METADATA_COMPUTE_LIMIT)]

        merge_plan = MergePlan(
            dedup_key="{}_{}_{}_{}".format("NSX_I", self.detector_name().name, anomaly_subtype.name, compute_id),
            metadata_key_merge_operation={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): MergePolicy.SET_UNION_WITH_MAX,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): MergePolicy.KEEP_ORIGINAL,
            },
        )

        viz_filter = VizFilter(
            key=VizFilterOption.VM_ID,
            values=potential_vms_responded_today + [compute_id],
            condition=Condition.OR,
            criteria=Criteria.EQ,
        )

        event_detail = self._create_event_detail(
            anomaly_subtype,
            num_new_vm_responses,
            len(potential_vms_responded_today),
            compute_row["request_received_count"],
            compute_row["response_count"],
            response_counts,
            responded_vm_counts,
            counts_labels,
        )

        event_metadata = EventMetadata(
            kv_store={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): ListString([compute_id]),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): ListString(
                    potential_vms_responded_today
                ),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_HOTSPOTS): ListString([compute_id]),
                EventKey.EVENT_METADATA_REPORT_KEY: JsonValue(json.dumps(event_detail.to_dict())),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): ListString(
                    [compute_to_ip_lookup[compute_id] if compute_id in compute_to_ip_lookup else ""]
                ),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): ListString(
                    [
                        compute_to_ip_lookup[compute]
                        for compute in potential_vms_responded_today
                        if compute in compute_to_ip_lookup
                    ]
                ),
                EventKey.EVENT_METADATA_VIZ_FILTER_KEY: JsonValue(viz_filter.to_json_str()),
            }
        )

        if LookupService.is_valid_ip_address(compute_id):
            # attacker is a public IP address
            compute_type = ComputeType.COMPUTE_TYPE_IP
            affected_compute_ip = compute_id
            ip_mask = IpMask.PUBLIC
        else:
            # attacker is a known VM
            compute_type = ComputeType.COMPUTE_TYPE_VM
            affected_compute_ip = compute_to_ip_lookup[compute_id] if compute_id in compute_to_ip_lookup else None
            ip_mask = IpMask.PRIVATE

        affected_compute = Compute(
            compute_type=compute_type,
            ip_address=affected_compute_ip,
            compute_id=compute_id,
            fqdn=None,
            site_id=site_id,
            ip_mask=ip_mask,
        )

        now_epoch_milli = int(self._now.timestamp() * 1000)
        data_flows = []
        for dst_compute_id in potential_vms_responded_today:
            if LookupService.is_valid_ip_address(dst_compute_id):
                dst_compute_type = ComputeType.COMPUTE_TYPE_IP
                dst_compute_ip = dst_compute_id
                dst_ip_mask = IpMask.PUBLIC
            else:
                dst_compute_type = ComputeType.COMPUTE_TYPE_VM
                dst_compute_ip = compute_to_ip_lookup.get(dst_compute_id)
                dst_ip_mask = IpMask.PRIVATE

            dst_compute = Compute(
                compute_type=dst_compute_type,
                ip_address=dst_compute_ip,
                compute_id=dst_compute_id,
                fqdn=None,
                site_id=site_id,
                ip_mask=dst_ip_mask,
            )

            data_flows.append(
                DataFlow(
                    src_compute=affected_compute,
                    dst_compute=dst_compute,
                    # We don't have transport protocol information
                    transport_protocols=None,
                    # We don't have application protocol information
                    application_protocols=None,
                    # We don't have server port information
                    server_ports=None,
                    # We don't have start and end time of the flows, use detection time
                    start_time=now_epoch_milli,
                    end_time=now_epoch_milli,
                )
            )

        event_info = EventInfo(
            version="v1.1",
            detector_name=self.detector_name(),
            detection_start_time_epoch_milli=now_epoch_milli,
            detection_end_time_epoch_milli=now_epoch_milli,
            score=Score(round(compute_row["confidence"]), severity=20),
            event_metadata=event_metadata,
            merge_plan=merge_plan,
            mitre_tactic=Tactic.CREDENTIAL_ACCESS,
            mitre_technique=Technique.LLMNR_NBTNS_POISONING,
            silent=compute_row["is_silent"],
            site_ids=[site_id],
            org_id=Config.ORG_ID,
            affected_compute=affected_compute,
            is_attacker=True,
            is_client=True,
            data_flows=data_flows,
        )

        return event_info

    def _create_event_detail(
        self,
        anomaly_sub_type: AnomalySubtype,
        num_new_vm_responses: int,
        num_vms_responded: int,
        num_requests: float,
        num_responses: float,
        response_counts: list[float],
        responded_vm_counts: list[int],
        counts_labels: list[pd.Timestamp],
    ) -> NameResolutionReport:
        counts_labels_as_str = [str(int(label.timestamp() * 1000)) for label in counts_labels]

        # data = baseline + current data ["response_count"]
        # labels = timestamps sorted by time
        mean, std = self._mean_and_std_without_outliers(response_counts)
        response_count = Series(
            data=[float(count) for count in response_counts], labels=counts_labels_as_str, mean=mean, stddev=std
        )

        # baseline + current data ["response_vms"].str.len
        # labels = timestamps sorted by time
        mean, std = self._mean_and_std_without_outliers(responded_vm_counts)
        vm_responded_count = Series(
            data=[float(count) for count in responded_vm_counts], labels=counts_labels_as_str, mean=mean, stddev=std
        )

        return NameResolutionReport(
            protocol=anomaly_sub_type.name,
            number_of_new_vms_responded=num_new_vm_responses,
            number_of_vms_responded=num_vms_responded,
            number_of_requests_received=int(num_requests),
            number_of_responses_sent=int(num_responses),
            response_count=response_count,
            vm_responded_count=vm_responded_count,
            report_type=self.detector_name().name,
            viz_filter_elements=None,
        )

    def _mean_and_std_without_outliers(self, values: list[int] | list[float]) -> tuple[float, float]:
        """returns (mean, std) of the given list of values with the outliers removed"""
        values_series = pd.Series(values)
        mean = values_series.mean()
        std = values_series.std()
        if std == 0:
            return (values_series.mean(), values_series.std())

        upper_bound = mean + (LlmnrNbtns.DEFAULT_STD * std)
        lower_bound = mean - (LlmnrNbtns.DEFAULT_STD * std)
        values_series = values_series[(values_series > lower_bound) & (values_series < upper_bound)]

        return (values_series.mean(), values_series.std())

    def _get_detector_run_status(
        self, status_label: DetectorStatusLabel, site_id: str, last_successful_run_time: int, message: str = ""
    ) -> DetectorRunStatus:
        LlmnrNbtns.logger.debug(f"Setting detector run status for {self.detector_name().name}")
        max_baseline_duration_days = self._baseline_service.get_max_baseline_duration_days(
            self.detector_name().name, Config.ORG_ID, site_id
        )
        remaining_baseline_duration_days = max(int(Config.MIN_BASELINE_DURATION_DAYS) - max_baseline_duration_days, 0)
        status_label = (
            DetectorStatusLabel.NOT_ENOUGH_BASELINE
            if (status_label == DetectorStatusLabel.SUCCESS and remaining_baseline_duration_days > 0)
            else status_label
        )

        return DetectorRunStatus(
            status_label=status_label,
            last_successful_run_time=last_successful_run_time,
            max_baseline_duration_days=max_baseline_duration_days,
            remaining_baseline_duration_days=remaining_baseline_duration_days,
            message=message,
        )
