import io
import ipaddress
import json
import pickle
import unittest
from datetime import datetime, timedelta
from random import randint
from unittest.mock import Mock

import numpy as np
import pandas as pd
from bloom_filter2 import BloomFilter

from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import Interval, PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.detector.core import DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import (
    ComputeType,
    IpMask,
)
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    AnomalyExclusionSetting,
    AnomalyExclusionSettings,
    DetectorRunStatus,
    DetectorStatusLabel,
    ExcludedItemsInfo,
    ExclusionLabel,
    ExclusionTypeInfo,
    UserConfigInfo,
)
from llmnrnbtns.detector.config import Config
from llmnrnbtns.detector.core import (
    AnomalySubtype,
    BaselineData,
    LlmnrNbtns,
    PartialBaselineData,
)


class TestCore(unittest.TestCase):
    def setUp(self):
        self._mock_baseline_service = Mock(BaselineService)
        self._mock_checkpoint_service = Mock(CheckpointService)
        self._mock_user_config_service = Mock(UserConfigService)
        self._mock_event_service = Mock(EventService)
        self._mock_flow_service = Mock(FlowService)
        self._mock_lookup_service = Mock(LookupService)
        self.detector = LlmnrNbtns(
            self._mock_baseline_service,
            self._mock_checkpoint_service,
            self._mock_user_config_service,
            self._mock_event_service,
            self._mock_flow_service,
            self._mock_lookup_service,
            datetime.now(),
        )

    def test_detector_name(self):
        self.assertEqual(DetectorType.LLMNR_NBTNS, self.detector.detector_name())

    def test_get_llmnr_partial_baseline_data(self):
        self._mock_flow_service.group_by = test_get_llmnr_partial_baseline_data_mock_query

        res = self.detector._get_llmnr_partial_baseline_data(Interval(datetime.now(), datetime.now()), "default")
        res_df = pd.read_parquet(res)

        self.assertTrue(pd.isnull(res_df.loc["vm11"]["response_count"]))
        self.assertTrue(pd.isnull(res_df.loc["vm12"]["response_count"]))
        self.assertTrue(pd.isnull(res_df.loc["vm13"]["response_count"]))
        self.assertTrue(pd.isnull(res_df.loc["vm14"]["response_count"]))
        self.assertTrue(pd.isnull(res_df.loc["vm3043"]["response_count"]))
        self.assertTrue(pd.isnull(res_df.loc["vm3090"]["response_count"]))
        self.assertTrue(pd.isnull(res_df.loc["vm3942"]["response_count"]))

        self.assertTrue(pd.isnull(res_df.loc["vm11"]["response_vms"]))
        self.assertTrue(pd.isnull(res_df.loc["vm12"]["response_vms"]))
        self.assertTrue(pd.isnull(res_df.loc["vm13"]["response_vms"]))
        self.assertTrue(pd.isnull(res_df.loc["vm14"]["response_vms"]))
        self.assertTrue(pd.isnull(res_df.loc["vm3043"]["response_vms"]))
        self.assertTrue(pd.isnull(res_df.loc["vm3090"]["response_vms"]))
        self.assertTrue(pd.isnull(res_df.loc["vm3942"]["response_vms"]))

        self.assertEqual(440, res_df.loc["vm9"]["request_received_count"])
        self.assertEqual(53, res_df.loc["vm9"]["response_count"])
        self.assertEqual(["vm8"], res_df.loc["vm9"]["response_vms"])

        self.assertEqual(5, res_df.loc["vm7"]["request_received_count"])
        self.assertEqual(924, res_df.loc["vm7"]["response_count"])
        self.assertEqual(["vm11", "vm12", "vm13"], list(res_df.loc["vm7"]["response_vms"]))

    def test_get_nbtns_partial_baseline_data(self):
        self._mock_flow_service.group_by = test_get_nbtns_partial_baseline_data

        res = self.detector._get_nbtns_partial_baseline_data(Interval(datetime.now(), datetime.now()), "default")

        res_df = pd.read_parquet(res)

        self.assertEqual(["vm1"], list(res_df.loc["vm10"]["response_vms"]))
        self.assertEqual(73, res_df.loc["vm10"]["response_count"])
        self.assertEqual(52, res_df.loc["vm10"]["request_received_count"])

    def test_combine_baseline_objects(self):
        intervals = []
        for i in range(1, 10, 2):
            interval = Interval(datetime(2023, 1, i), datetime(2023, 1, i + 1))
            intervals.append(interval)

        baselines = []
        for i in range(5):
            llmnr_partial_baseline = []
            nbtns_partial_baseline = []
            for j in range(10):
                if j % 3 == 0:
                    llmnr_partial_baseline.append([f"vm{j}", randint(0, 999), np.NaN, None])
                    nbtns_partial_baseline.append([f"vm{j}", None, np.NaN, 0.0 + randint(0, 999)])

                elif j % 4 == 0:
                    llmnr_partial_baseline.append([f"vm{j}", randint(0, 999), np.NaN, None])
                    nbtns_partial_baseline.append([f"vm{j}", ["vm1, vm2"], 0.0 + randint(0, 999), np.NaN])

                else:
                    llmnr_partial_baseline.append([f"vm{j}", randint(0, 999), 0.0 + randint(0, 999), ["vm10", "vm20"]])
                    nbtns_partial_baseline.append(
                        [f"vm{j}", ["vm1, vm2"], 0.0 + randint(0, 999), 0.0 + randint(0, 999)]
                    )

            llmnr_partial_baseline_df = pd.DataFrame(
                llmnr_partial_baseline, columns=["vm_id", "request_received_count", "response_count", "response_vms"]
            )
            llmnr_partial_baseline_df = llmnr_partial_baseline_df.set_index("vm_id")

            nbtns_partial_baseline_df = pd.DataFrame(
                nbtns_partial_baseline, columns=["vm_id", "response_vms", "response_count", "response_received_count"]
            )
            nbtns_partial_baseline_df = nbtns_partial_baseline_df.set_index("vm_id")

            llmnr_buffer = io.BytesIO()
            nbtns_buffer = io.BytesIO()
            llmnr_partial_baseline_df.to_parquet(llmnr_buffer)
            nbtns_partial_baseline_df.to_parquet(nbtns_buffer)
            partial_baseline_data = PartialBaselineData(llmnr_buffer, nbtns_buffer)
            buffer = io.BytesIO()
            pickle.dump(partial_baseline_data, buffer)
            partial_baseline = PartialBaseline(intervals[i], buffer)
            baselines.append(partial_baseline)

        result = self.detector._combine_baseline_objects(baselines)
        self.assertIsNotNone(result)
        if result is not None:
            result.seek(0)
        baseline_data: BaselineData = pickle.load(result)  # type: ignore
        llmnr_result = pd.read_parquet(baseline_data.llmnr_data)
        nbtns_result = pd.read_parquet(baseline_data.nbtns_data)

        self.assertEqual(datetime(2023, 1, 1), llmnr_result.loc["vm1"].index.min())
        self.assertEqual(datetime(2023, 1, 1), nbtns_result.loc["vm1"].index.min())
        self.assertEqual(datetime(2023, 1, 9), llmnr_result.loc["vm9"].index.max())
        self.assertEqual(datetime(2023, 1, 9), nbtns_result.loc["vm9"].index.max())

        self.assertTrue("vm1:vm10" in baseline_data.llmnr_datacenter_data_per_day[datetime(2023, 1, 1)])
        self.assertTrue("vm1:vm20" in baseline_data.llmnr_datacenter_data_per_day[datetime(2023, 1, 1)])

    def _assert_valid_compute(self, compute):
        if compute.compute_type == ComputeType.COMPUTE_TYPE_VM:
            self.assertIsNotNone(compute.compute_id)

            # Compute.compute_d should be a UUID (here in the form vm-XYZ), but should not parse
            # into a valid IP address
            with self.assertRaises(ValueError):
                ipaddress.ip_address(compute.compute_id)
            self.assertEqual(compute.ip_mask, IpMask.PRIVATE)
        else:
            self.assertIsNotNone(compute.ip_address)
            # This shouldn't fail
            ipaddress.ip_address(compute.ip_address)
            self.assertEqual(compute.ip_mask, IpMask.PUBLIC)

    def test_detect_events_on(self):
        exclusion_vm = AnomalyExclusionSetting(
            label=ExclusionLabel.DESTINATION,
            type=ExclusionTypeInfo.VM,
            excluded_items_info=ExcludedItemsInfo(["vm10", "vm11", "vm14"]),
        )

        exclusions = AnomalyExclusionSettings([exclusion_vm])
        mock_sensitivity = Mock()
        mock_sensitivity_inner = Mock()
        mock_sensitivity.get_sensitivity_meter_setting = lambda: mock_sensitivity_inner
        mock_sensitivity_inner.sensitivity = 50
        user_config = UserConfigInfo(
            "site 1",
            "",
            "",
            "",
            0,
            True,
            False,
            "",
            False,
            50,
            [],
            [],
            mock_sensitivity,
            exclusions,
            DetectorRunStatus(DetectorStatusLabel.NOT_SCHEDULED),
        )

        computes_detecting_on = [
            ["vm1", ["vm10"], 124, 45],
            ["vm2", ["vm10", "vm4"], 0, 0],
            ["vm3", ["vm1", "vm9", "vm2"], 1.23, 4.51],
            ["vm4", None, np.NaN, 23.0],
            ["vm5", None, np.NaN, 23],
            ["vm6", ["vm1", "vm2", "vm3", "vm4", "vm5", "vm7", "vm8", "vm9"], 1688.0, np.NaN],
            ["vm7", ["vm100", "vm200", "vm300", "vm400", "vm500"], 421, 150],
            ["vm8", ["vm1", "vm2", "vm3"], 56, 3],
            ["vm9", ["vm2", "vm3", "vm4", "vm5", "vm6", "vm7", "vm8"], 1600, 96],
            ["vm10", ["vm6"], 150, 0],
            ["vm20", ["vm6", "vm7", "vm8", "vm9"], 150, np.NaN],
            ["vm30", ["vm6"], 4, 43],
            ["vm40", ["vm6"], 4, 43],
        ]

        computes_detecting_on_df = pd.DataFrame(
            computes_detecting_on, columns=["vm_id", "response_vms", "response_count", "request_received_count"]
        ).set_index("vm_id")

        baseline_data = [
            ["vm1", datetime(year=2023, month=1, day=1), ["vm10"], 36, 345],
            ["vm1", datetime(year=2023, month=1, day=2), ["vm10", "vm2"], 42, 89],
            ["vm1", datetime(year=2023, month=1, day=3), ["vm10", "vm3"], 42, 89],
            ["vm1", datetime(year=2023, month=1, day=4), ["vm10"], 543, 94],
            ["vm2", datetime(year=2023, month=1, day=1), ["vm10"], 48, 83],
            ["vm2", datetime(year=2023, month=1, day=2), ["vm10"], 20, 99],
            ["vm2", datetime(year=2023, month=1, day=3), ["vm10"], 55, 34],
            ["vm2", datetime(year=2023, month=1, day=4), ["vm10"], 28, 36],
            ["vm3", datetime(year=2023, month=1, day=1), ["vm10", "vm1"], 42, 89],
            ["vm3", datetime(year=2023, month=1, day=2), ["vm10", "vm2"], 42, 89],
            ["vm3", datetime(year=2023, month=1, day=3), ["vm10", "vm6"], 42, 89],
            ["vm3", datetime(year=2023, month=1, day=4), ["vm10"], 72, 544],
            ["vm4", datetime(year=2023, month=1, day=1), ["vm10"], 562, 433],
            ["vm4", datetime(year=2023, month=1, day=2), ["vm10", "vm1", "vm2"], 42, 89],
            ["vm4", datetime(year=2023, month=1, day=3), None, np.NaN, 672],
            ["vm4", datetime(year=2023, month=1, day=4), ["vm10"], 234, 36],
            ["vm5", datetime(year=2023, month=1, day=1), ["vm10"], 6, 89],
            ["vm5", datetime(year=2023, month=1, day=2), ["vm10"], 34, 34],
            ["vm5", datetime(year=2023, month=1, day=3), ["vm10"], 62, 44],
            ["vm5", datetime(year=2023, month=1, day=4), ["vm10"], 72, 234],
            ["vm6", datetime(year=2023, month=1, day=1), None, np.NaN, 869],
            ["vm6", datetime(year=2023, month=1, day=2), None, np.NaN, 89],
            ["vm6", datetime(year=2023, month=1, day=3), None, np.NaN, 62],
            ["vm6", datetime(year=2023, month=1, day=4), None, np.NaN, 67],
            ["vm7", datetime(year=2023, month=1, day=1), ["vm1"], 63, 27],
            ["vm7", datetime(year=2023, month=1, day=2), ["vm10", "vm4", "vm2"], 42, 24],
            ["vm7", datetime(year=2023, month=1, day=3), ["vm2"], 34, 33],
            ["vm7", datetime(year=2023, month=1, day=4), ["vm1"], 63, 14],
            ["vm8", datetime(year=2023, month=1, day=1), ["vm10"], 6, 673],
            ["vm8", datetime(year=2023, month=1, day=2), [], 54, 345],
            ["vm8", datetime(year=2023, month=1, day=3), ["vm10"], 34, 64],
            ["vm8", datetime(year=2023, month=1, day=4), ["vm10"], 462, 26],
            ["vm9", datetime(year=2023, month=1, day=1), [], 72, 434],
            ["vm9", datetime(year=2023, month=1, day=2), [], 43, 44],
            ["vm9", datetime.now() - timedelta(days=1), [], 634, 72],
            ["vm9", datetime.now() - timedelta(hours=1), [], 42, 89],
            ["vm10", datetime.now() - timedelta(hours=1), ["vm10"], 42, 89],
            ["vm10", datetime.now() - timedelta(hours=1), ["vm10"], 42, 89],
            ["vm10", datetime.now() - timedelta(hours=1), ["vm10"], 42, 89],
            ["vm10", datetime.now() - timedelta(hours=1), ["vm10"], 42, 89],
            ["vm100", datetime(year=2023, month=1, day=1), ["vm10"], 634, 46],
            ["vm100", datetime(year=2023, month=1, day=1), ["vm10"], 24, 43],
            ["vm100", datetime(year=2023, month=1, day=1), ["vm10"], 75, 644],
            ["vm100", datetime(year=2023, month=1, day=1), ["vm10"], 43, 34],
            ["vm200", datetime(year=2023, month=1, day=1), ["vm10"], 63, 34],
            ["vm200", datetime(year=2023, month=1, day=1), ["vm10"], 63, 89],
            ["vm200", datetime(year=2023, month=1, day=1), ["vm10"], 734, 673],
            ["vm200", datetime(year=2023, month=1, day=1), ["vm10"], 42, 44],
            ["vm300", datetime(year=2023, month=1, day=1), ["vm10"], 62, 32],
            ["vm300", datetime(year=2023, month=1, day=1), ["vm10"], 63, 24],
            ["vm300", datetime(year=2023, month=1, day=1), ["vm10"], 45, 345],
            ["vm300", datetime(year=2023, month=1, day=1), ["vm10"], 63, 34],
            ["vm300", datetime(year=2023, month=1, day=1), ["vm10"], 63, 34],
        ]

        baseline_data_df = pd.DataFrame(
            baseline_data, columns=["vm_id", "start_time", "response_vms", "response_count", "request_received_count"]
        ).set_index(["vm_id", "start_time"])

        datacenter_data: BloomFilter = BloomFilter(max_elements=100000, error_rate=0.001)
        datacenter_data.add("vm1:vm2")
        datacenter_data.add("vm2:vm1")
        datacenter_data.add("vm3:vm1")
        datacenter_data.add("vm1:vm3")
        datacenter_data_per_data = {datetime.now() - timedelta(days=1): datacenter_data}

        self._mock_flow_service.group_by = test_detect_events_on_mock_query

        compute_to_ip = {
            "vm1": "192.168.0.1",
            "vm2": "192.168.0.2",
            "vm3": "192.168.0.3",
            "vm4": "192.168.0.4",
            "vm5": "192.168.0.5",
            "vm100": "192.168.0.100",
            "vm200": "192.168.0.200",
        }

        self._mock_lookup_service.computes_to_ip_lookup = lambda _site_id, _interval: compute_to_ip

        events = [
            _
            for _ in self.detector._detect_events_on(
                user_config, computes_detecting_on_df, baseline_data_df, datacenter_data_per_data, AnomalySubtype.LLMNR
            )
        ]

        self.assertEqual(73, events[2].score.confidence)
        self.assertEqual(["vm7"], events[2].event_metadata.kv_store["METADATA_KEY_HOTSPOTS"].list_string_value)  # type: ignore
        for event in events:
            event_metadata_dict = json.loads(event.event_metadata.kv_store["REPORT"].json_value)  # type: ignore
            response_counts = event_metadata_dict["response_count"]
            response_counts_data = response_counts["data"]
            response_counts_labels = response_counts["labels"]

            # Basic data flow check, let's make sure we have a valid set of endpoints
            for data_flow in event.data_flows or []:
                self._assert_valid_compute(data_flow.src_compute)
                self._assert_valid_compute(data_flow.dst_compute)

            # self.assertGreater(response_counts_data[-1], response_counts_data[-2])
            # self.assertGreater(response_counts_labels[-1], response_counts_labels[-2])

    def test_create_event(self):
        result = self.detector._create_event(
            "test_compute_id",
            pd.DataFrame(
                [[140, 200, 50, False]], columns=["request_received_count", "response_count", "confidence", "is_silent"]
            ),  # type: ignore
            AnomalySubtype.LLMNR,
            [f"{i} compute" for i in range(100)],
            12,
            [i for i in range(100)],
            [i for i in range(1000)],
            [pd.Timestamp.now() for _ in range(100)],
            "test site id",
            {f"{i} compute": f"{i} ip addr" for i in range(100)},
        )

        self.assertEqual(
            100,
            len(result.event_metadata.kv_store.get("METADATA_KEY_DST_COMPUTES").list_string_value),  # type: ignore
        )

        self.assertEqual(
            100,
            len(result.event_metadata.kv_store.get("METADATA_KEY_DST_IPS").list_string_value),  # type: ignore
        )

        # now, apply the limit
        original_compute_limit = Config.EVENT_METADATA_COMPUTE_LIMIT
        Config.EVENT_METADATA_COMPUTE_LIMIT = "2"
        result = self.detector._create_event(
            "test_compute_id",
            pd.DataFrame(
                [[140, 200, 50, False]], columns=["request_received_count", "response_count", "confidence", "is_silent"]
            ),  # type: ignore
            AnomalySubtype.LLMNR,
            [f"{i} compute" for i in range(100)],
            12,
            [i for i in range(100)],
            [i for i in range(1000)],
            [pd.Timestamp.now() for _ in range(100)],
            "test site id",
            {f"{i} compute": f"{i} ip addr" for i in range(100)},
        )

        self.assertEqual(
            ["0 compute", "1 compute"],
            result.event_metadata.kv_store.get("METADATA_KEY_DST_COMPUTES").list_string_value,  # type: ignore
        )

        self.assertEqual(
            ["0 ip addr", "1 ip addr"],
            result.event_metadata.kv_store.get("METADATA_KEY_DST_IPS").list_string_value,  # type: ignore
        )

        # reset compute limit for other tests
        Config.EVENT_METADATA_COMPUTE_LIMIT = original_compute_limit


def test_detect_events_on_mock_query(query: FlowGroupByQuery) -> pd.DataFrame:
    result = [
        ["vm1", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm2", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm3", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm4", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm5", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm6", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm7", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm8", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm9", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm10", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm11", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm12", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm13", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm100", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm200", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm300", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm400", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm500", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
        ["vm600", datetime(year=2023, month=1, day=1), datetime(year=2023, month=1, day=1)],
    ]

    result_df = pd.DataFrame(result, columns=["srcVmId", "lastStartTime", "timestamp"])
    result_df = pd.DataFrame(result, columns=["srcVmId", "lastStartTime", "timestamp"])

    return result_df


def test_get_llmnr_partial_baseline_data_mock_query(query: FlowGroupByQuery) -> pd.DataFrame:
    result_count = [
        ["vm1", "vm10", 0, datetime(year=2023, month=1, day=1)],
        ["vm2", "vm10", 5, datetime(year=2023, month=1, day=1)],
        ["vm3", "vm20", 10, datetime(year=2023, month=1, day=1)],
        ["vm4", "vm10", 1, datetime(year=2023, month=1, day=1)],
        ["vm5", "vm20", 15, datetime(year=2023, month=1, day=1)],
        ["vm10", "vm40", 0, datetime(year=2023, month=1, day=2)],
        ["vm11", "vm70", 5, datetime(year=2023, month=1, day=2)],
        ["vm12", "vm70", 20, datetime(year=2023, month=1, day=2)],
        ["vm13", "vm70", 6, datetime(year=2023, month=1, day=2)],
        ["vm14", "vm60", 30, datetime(year=2023, month=1, day=2)],
        ["vm6", "vm50", 0, datetime(year=2023, month=1, day=3)],
        ["vm7", "vm40", 5, datetime(year=2023, month=1, day=3)],
        ["vm8", "vm90", 30, datetime(year=2023, month=1, day=3)],
        ["vm9", "vm30", 400, datetime(year=2023, month=1, day=3)],
        ["vm9", "vm80", 40, datetime(year=2023, month=1, day=4)],
        ["vm3090", "vm5", 42, datetime(year=2023, month=1, day=4)],
        ["vm3043", "vm2", 64, datetime(year=2023, month=1, day=4)],
        ["vm3942", "vm1", 2, datetime(year=2023, month=1, day=4)],
    ]

    result_count_df = pd.DataFrame(result_count, columns=["dstVmId", "srcVmId", "count", "timestamp"])

    result_source_port = [
        ["vm1", "0", datetime(year=2023, month=1, day=1)],
        ["vm2", "5", datetime(year=2023, month=1, day=1)],
        ["vm3", "10", datetime(year=2023, month=1, day=1)],
        ["vm4", "1", datetime(year=2023, month=1, day=1)],
        ["vm5", "15", datetime(year=2023, month=1, day=1)],
        ["vm10", "0", datetime(year=2023, month=1, day=2)],
        ["vm11", "5", datetime(year=2023, month=1, day=2)],
        ["vm12", "20", datetime(year=2023, month=1, day=2)],
        ["vm13", "6", datetime(year=2023, month=1, day=2)],
        ["vm14", "30", datetime(year=2023, month=1, day=2)],
        ["vm6", "0", datetime(year=2023, month=1, day=3)],
        ["vm7", "5", datetime(year=2023, month=1, day=3)],
        ["vm8", "30", datetime(year=2023, month=1, day=3)],
        ["vm9", "400", datetime(year=2023, month=1, day=3)],
        ["vm9", "40", datetime(year=2023, month=1, day=4)],
        ["vm2090", "40", datetime(year=2023, month=1, day=4)],
        ["vm2043", "32", datetime(year=2023, month=1, day=4)],
        ["vm2942", "21", datetime(year=2023, month=1, day=4)],
    ]

    result_source_port_df = pd.DataFrame(result_source_port, columns=["srcVmId", "originalSrcPorts", "timestamp"])

    response = [
        ["vm1", "vm10", "0", 342, datetime(year=2023, month=1, day=1)],
        ["vm2", "vm1", "0", 532, datetime(year=2023, month=1, day=1)],
        ["vm3", "vm2", "10", 75, datetime(year=2023, month=1, day=1)],
        ["vm4", "vm1", "1", 345, datetime(year=2023, month=1, day=1)],
        ["vm5", "vm2", "15", 754, datetime(year=2023, month=1, day=1)],
        ["vm10", "vm4", "0", 332, datetime(year=2023, month=1, day=2)],
        ["vm11", "vm7", "5", 97, datetime(year=2023, month=1, day=2)],
        ["vm12", "vm7", "20", 774, datetime(year=2023, month=1, day=2)],
        ["vm13", "vm7", "6", 53, datetime(year=2023, month=1, day=2)],
        ["vm14", "vm6", "30", 64, datetime(year=2023, month=1, day=2)],
        ["vm6", "vm5", "0", 65, datetime(year=2023, month=1, day=3)],
        ["vm7", "vm4", "5", 32, datetime(year=2023, month=1, day=3)],
        ["vm8", "vm9", "30", 53, datetime(year=2023, month=1, day=3)],
        ["vm9", "vm3", "400", 674, datetime(year=2023, month=1, day=3)],
        ["vm9", "vm8", "40", 246, datetime(year=2023, month=1, day=4)],
        ["vm1090", "vm8", "40", 246, datetime(year=2023, month=1, day=4)],
        ["vm1043", "vm8", "40", 246, datetime(year=2023, month=1, day=4)],
        ["vm1942", "vm8", "40", 246, datetime(year=2023, month=1, day=4)],
    ]

    response_df = pd.DataFrame(response, columns=["dstVmId", "srcVmId", "dstPort", "count", "timestamp"])

    dimensions = []
    for dimension in query.dimensions:
        dimensions.append(dimension._dimension)  # type: ignore

    if "dstPort" in dimensions:
        return response_df

    if "originalSrcPorts" in dimensions:
        return result_source_port_df

    else:
        return result_count_df


def test_get_nbtns_partial_baseline_data(query: FlowGroupByQuery) -> pd.DataFrame:
    response = [
        ["vm1", "vm10", "192.168.0.1", 12, 73, datetime(year=2023, month=1, day=1)],
        ["vm2", "vm1", "192.168.0.2", 53, 24, datetime(year=2023, month=1, day=1)],
        ["vm3", "vm2", "192.168.0.3", 6, 74, datetime(year=2023, month=1, day=1)],
        ["vm4", "vm1", "192.168.0.4", 53, 254, datetime(year=2023, month=1, day=1)],
        ["vm5", "vm2", "192.168.0.5", 32, 634, datetime(year=2023, month=1, day=1)],
        ["vm10", "vm4", "192.168.0.10", 52, 72, datetime(year=2023, month=1, day=2)],
        ["vm11", "vm7", "192.168.0.11", 3, 54, datetime(year=2023, month=1, day=2)],
        ["vm12", "vm7", "192.168.0.12", 67, 33, datetime(year=2023, month=1, day=2)],
        ["vm13", "vm7", "192.168.0.13", 3, 32, datetime(year=2023, month=1, day=2)],
        ["vm14", "vm6", "192.168.0.14", 1, 11, datetime(year=2023, month=1, day=2)],
        ["vm6", "vm5", "192.168.0.6", 64, 13, datetime(year=2023, month=1, day=3)],
        ["vm7", "vm4", "192.168.0.7", 3, 164, datetime(year=2023, month=1, day=3)],
        ["vm8", "vm9", "192.168.0.8", 53, 366, datetime(year=2023, month=1, day=3)],
        ["vm9", "vm3", "192.168.0.9", 25, 334, datetime(year=2023, month=1, day=3)],
        ["vm9", "vm8", "192.168.0.9", 63, 65, datetime(year=2023, month=1, day=4)],
        ["vm1090", "vm8", "192.168.0.1090", 25, 34, datetime(year=2023, month=1, day=4)],
        ["vm1043", "vm8", "192.168.0.1043", 673, 63, datetime(year=2023, month=1, day=4)],
        ["vm1942", "vm8", "192.168.0.1942", 5, 22, datetime(year=2023, month=1, day=4)],
    ]

    flow_df = pd.DataFrame(
        response, columns=["dstVmId", "srcVmId", "dstIps", "dstFwAllowedSum", "totalFlowCount", "timestamp"]
    )

    if "dstIps" in query.dimensions:
        return flow_df

    ips = [
        ["192.168.0.1", datetime(year=2023, month=1, day=1)],
        ["192.168.0.2", datetime(year=2023, month=1, day=1)],
        ["192.168.0.3", datetime(year=2023, month=1, day=1)],
        ["192.168.0.4", datetime(year=2023, month=1, day=1)],
        ["192.168.0.5", datetime(year=2023, month=1, day=1)],
        ["192.168.0.6", datetime(year=2023, month=1, day=2)],
        ["192.168.0.7", datetime(year=2023, month=1, day=2)],
        ["192.168.0.8", datetime(year=2023, month=1, day=2)],
        ["192.168.0.45", datetime(year=2023, month=1, day=3)],
    ]

    ips_df = pd.DataFrame(ips, columns=["ip_addresses", "timestamp"])

    return ips_df


def test_get_detector_status(self):
    self.mock_baseline_service = Mock(BaselineService())
    self.mock_baseline_service.get_max_baseline_duration_days = lambda _x, _y, _z: 2
    original_baseline_service = self.detector._baseline_service
    self.detector._baseline_service = self.mock_baseline_service

    # Not enough baseline case
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.SUCCESS, "123", 1234, "success")
    self.assertEqual(DetectorStatusLabel.NOT_ENOUGH_BASELINE, detector_run_status.status_label)

    # FAILURE case
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.FAILURE, "123", 1234, "failed")
    self.assertEqual(DetectorStatusLabel.FAILURE, detector_run_status.status_label)

    # SUCCESS case - baseline size > min required baseline
    self.mock_baseline_service.get_max_baseline_duration_days = lambda _x, _y, _z: 7
    detector_run_status = self.detector._get_detetor_run_status(DetectorStatusLabel.SUCCESS, "123", 1234, "success")
    self.assertEqual(DetectorStatusLabel.SUCCESS, detector_run_status.status_label)

    self.detector._baseline_service = original_baseline_service
